{$A+ Align data on word boundary }
{$F- Don't force FAR calls       }
{$B- Short circuit boolean eval  }
{$V- Relax string checking       }

{$I- Enable System.IOResult error checks }
{$N+,E+ Allow for software 80x87 }
{$R- Range checking off }
{$S- Stack checking off }

{$L+ Enable gen. of local symbols }
{$D+ Enable debug information     }
{$O- No overlays                  }

{$M  4096,0, 655360} { Max-Stack Low-heap High-heap }

PROGRAM LNGswap; { (C) Copyright Adolfo Di Mare, 1994 }

{ NOTICE
  Use this programs at your own risk. The author
  reserves all comercial rights. This program, its
  source code and documentation have been put in the
  public domain.                                     }

{ USAGE:
  LNGswap lngfile[.ext] [LNG(language)]

  Switches all the annotated lines in the files
  denoted by lngfile[.ext] so that the current
  language is the one specified in the command
  line.

  For this, LNGswap moves all lines in every line
  group leaving in the first position the one
  that has the LNG(language) annotation.

  - lngfile is obtained running program LNGrep, and
    editing the resulting file.

  - LNGswap contains a moderate amount of error
    checking to protect the programmer from
    him/herself. However, when the line numbers
    for each group get messed up, the original
    source files can be damaged to no recognition.

  - Wildcards in lngfile[.ext] are processed
    correctly.

  - The original source files are retained with a
    .bak extention.

  - .bak file extensions are not processed.
                                                     }
{ EXAMPLE:
   C:> LNGswap *.prg LNG(spanish)

   In this case, all the annotations in the source
   files are moved around so that the line that
   leads each group of annotated lines the one with
   an Spanish annotation.                            }

{ PROGRAMMER                       Adolfo Di Mare    }

USES
  Dos, Crt, LngTools;


PROCEDURE WriteUSAGE;
{ RESULT
  Displays a help message. }
BEGIN { WriteUSAGE }
  WriteLn('LNGswap v', LngTools.LNG_version:1);
  WriteLn('USAGE: ');
  WriteLn('LNGswap lngfile[.ext] [LNG(language)]');
  WriteLn;
  WriteLn(
  'Switches all the annotated lines in the files');
  WriteLn(
  'denoted by lngfile[.ext] so that the current');
  WriteLn(
  'language is the one specified in the command');
  WriteLn(
  'line.');
END; { WriteUSAGE }


VAR
  SG : TString_Group;
  { SG is so big that it would blow up the run }
  { time stack if it were a local variable...  }

PROCEDURE Swap_File(                   { ADOLFO }
  {+} def_lng  : LNG_T;    { default annotation }
  {+} VAR dir  : Dos.DirStr;     { Dos.Fsplit() }
  {+} VAR fname: Dos.PathStr  { file to process }
);
{ RESULT
  "fname" is a program source file, written in
  any of the xBase dialects, that contains group
  of anoted lines.

  Each group of annotated lines is identified
  by Swap_File(). The line that has an annotation
  equal to "def_lng" is the one that gets
  written out as the first, non commented, line.
  All the others are commented out.

  - As a side effect, a backup copy of "fname"
    is generated.                               }

{ RESULT
  Reads the text file "fname", line by line,
  and identifies annotated group of lines, that
  correspond to different languages, and makes
  the first and non commented out line that
  which has a language annotation equal to
  "lng".

  - Only lines that already have an annotation
    are processed. If the need rises to process
    the other lines, then program LNGrep should
    be run first.                              }

{ REQUIRES
  The DOS file "fname" must exist.             }

VAR
  prgF : FILE_OF_CHAR_T; { file "fname" }
  newF : TEXT;

  str : Tokenized_T;

  str_lng: STRING[LNG_size];

  i : WORD;
  line_no : WORD; { current line in "fname" }

VAR
  ignore: Dos.DirStr;
  path  : Dos.PathStr;
  name  : Dos.NameStr;
  ext   : Dos.ExtStr;

BEGIN { Swap_File }

  IF IS_bak(fname) THEN BEGIN
    { avoid overwriting backup files }
    EXIT;
  END;

  { parse the file name }
  Dos.Fsplit(fname, ignore, name, ext);

  { create the .bak file }
  path := dir+name;
  Rename_to_BAK(path, ext);

  { open the two new files }
  Assign(prgF, path + '.bak');
  Reset(prgF);
  { the input file is fname.bak }

  { now create the new file, where   }
  { the annotations will be included }
  Assign(newF, dir + fname);
  ReWrite(newF); { old name, new file }

  { initialize the string group }
  SG.Init;

  line_no := 0;

  WHILE Get_Line(prgF, str.str) DO BEGIN

    INC(line_no);
    Tokenize(str.str, str.prs, str.com, str.lng);

    IF str.lng = '' THEN BEGIN
      { clear the previous group }
      SG.Write(newF, def_lng);

      { there is no annotation: just copy the line }
      WriteLn(newF, str.str);
    END

    ELSE IF str.lng = LNG_all THEN BEGIN
      { skip language independent strings }
      SG.Write(newF, def_lng);

      { copy the line }
      WriteLn(newF, str.str);
    END

    ELSE BEGIN
      { the line already has an LNG() annotation }

      IF (NOT str.com)      OR { not a comment?  }
         SG.Empty           OR { not in a group? }
         SG.IS_Member(str.lng) { already there?  }
      THEN BEGIN

        { a new group begins: flush the old one }
        SG.Write(newF, def_lng);

        { now decide whether to begin a new group }
        IF NOT str.com THEN BEGIN
          SG.Insert(str, line_no);
        END
        ELSE BEGIN
          { copy the line }
          WriteLn(newF, str.str);
        END;

      END
      ELSE BEGIN
      {
        (str.com)      AND          it's a comment
        (NOT SG.Empty) AND          there is a group
        (NOT SG.IS_Member(str.lng)) not in the group
      }
        { add the line to the group }
        SG.Insert(str, line_no);
      END;

    END;

  END; { WHILE }

  SG.Write(newF, def_lng);

  IF NOT Eof(prgF) THEN BEGIN
    Write  (newF, 'LNGrep: line too long in file ');
    WriteLn(newF,  fname, ' (', line_no, ')');
    WriteLn(newF, 'LNGrep: file scanning aborted!!!');
  END;

  SG.Done;
  Close(prgF);
  Close(newF);
END; { Swap_File }


TYPE
  Turbo_SearchRec = RECORD       { from Online Help }
    fill: ARRAY[1..21] OF BYTE;
    attr: BYTE;
    time: LONGINT;
    size: LONGINT;
    name: STRING[12];
  END;

  { String types, from ...\TURBO\DOC\Dos.int }
  Dos_ComStr  = STRING[127]; { Command line string }
  Dos_PathStr = STRING[79];  { File pathname string }
  Dos_DirStr  = STRING[67];  { Drive and dir. string }
  Dos_NameStr = STRING[8];   { File name string }
  Dos_ExtStr  = STRING[4];   { File extension string }


VAR
  def_lng  : LNG_T;
  ignore1,
  ignore2  : BOOLEAN;
  wildcard : STRING;

VAR
  dir : Dos.DirStr;
{ path: Dos.PathStr;  }
  name: Dos.NameStr;
  ext : Dos.ExtStr;

VAR
  srec   : Dos.SearchRec;

CONST
  {$IFNDEF VER55}   FName_S = SizeOf(srec.name);
  {$ELSE}           FName_S = 1+12;
  {$ENDIF !VER55}

TYPE
  Node_P = ^Node_T;
  Node_T = RECORD
    fname : STRING[FName_S];
    next  : Node_P;
  END;
VAR
  list,
  node,
  last   : Node_P;

BEGIN { LNGswap }
  { get the command line parameters }
  IF NOT Command_Line(
           def_lng, wildcard, ignore1, ignore2)
  THEN BEGIN
    WriteUSAGE;
    EXIT;
  END;

  { parse the file name }
  Dos.Fsplit(wildcard, dir, name, ext);

  { "dir" contains the drive/directory }
  { "dir" is the path to each file     }

  { create a list with all the file names }
  Dos.FindFirst(
    wildcard,
    Dos.Archive + Dos.ReadOnly,
    srec);
  list := NIL;

  WHILE Dos.DosError = 0 DO BEGIN

    NEW(node);
    node^.fname := srec.name;

    IF list = NIL THEN BEGIN
      list := node;   { empty list }
    END
    ELSE BEGIN
      last^.next := node; { append }
    END;

    last := node;

    Dos.FindNext(srec);

  END;
{
   DosError  Meaning
   
         2   File not found
         3   Path not found
         5   Access denied
         6   Invalid handle
         8   Not enough memory
        10   Invalid environment
        11   Invalid format
        18   No more files
}
  { finish up the list }
  IF list <> NIL THEN BEGIN
    last^.next := NIL;
  END;
{
  It seems unnecesary to create a list with all the
  file names. At first glance, it will appear that
  simply scanning the current directory with
  Dos.FindNext() is enough. This is not the case
  because procedure Swap_File() renames the source
  program file [.prg] to a .bak extention.

  To rename a file, DOS changes in place the directory
  entry.  When the new [.prg] file is created, this
  new directory entry will appear at the end of the
  directory. This means that Dos.FindNext() will find
  it again, and the file will be processed twice: once
  with the original file, and one with the new source.
  The following figure depicts this:

     Initial       Rename      Create        Final
   +---------+  +---------+  +---------+  +---------+
   |other.pas|  |other.pas|  |other.pas|  |other.pas|
   >short.prg<  >short.bak<  >short.bak<  >short.prg<
   |grep.com |  |grep.com |  |grep.com |  |grep.com |
   |         |  |         |  >short.prg<  >short.bak<
   +---------+  +---------+  +---------+  +---------+

  The second time that the .bak file is created, it
  will have the same contents as the source. The net
  result of this is that the final source file will be
  exactly equal to the .bak file, which means that no
  .bak copy of the file will remain in the directory.
  This is not what it is intended.

  This minor DOS bug (or virtue) can be overcomed only
  reading all directory entries before processing each
  file.

  - This bug made a previous verion of Swap_File() to
    leave the .bak file equal to the new source file,
    because the new source was processed again.

  - An infinity loop never ocurred while testing, but
    in theory it could happen.
}

  { process each file }
  WHILE list <> NIL DO BEGIN

    Swap_File(def_lng, dir, list^.fname);

    { cleanup dynamic memory }
    node := list;
    list := list^.next;
    DISPOSE(node);
  END;

END. { LNGswap }
